import re
from collections.abc import Callable, Sequence
from pathlib import Path

from presidio_analyzer import AnalyzerEngine

from pteredactyl.exceptions import MissingRegexRecogniserError
from pteredactyl.recognisers.pteredactyl_recogniser import (
    PTEREDACTYL_RECOGNISER_NAME,
    PteredactylRecogniser,
)
from pteredactyl.regex_check_functions import is_nhs_number

REGEX_ENTITIES = {
    # entity_type: (regex, check_function)
    "NHS_NUMBER": (r"\d(?:[\s-]?\d){9,}", is_nhs_number),
    "POSTCODE": (r"[A-Z]{1,2}[0-9R][0-9A-Z]?\s?[0-9][A-Z]{2}", None),
    "EMAIL_ADDRESS": (r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+", None),
}


def build_pteredactyl_recogniser(
    entity_type: str,
    regex: str | re.Pattern,
    check_function: Callable[..., bool] | None,
) -> PteredactylRecogniser:
    """
    Build a custom regex ptererecogniser for pteredactyl.

    Args:
        entity_type (str): The name of the entity to be recognised.
        regex (str or re.Pattern): The regular expression to match the entity.
        check_function (Callable, optional): A function to check if the matched string is a valid entity. Should take a single argument (the matched string) and return a boolean.

    Returns:
        PteredactylRecogniser: A custom presidio EntityRecognizer object.

    Example:
    >>> def check_soton_landline(input: str):
    ... cleaned = input.replace('-','').replace(' ','')
    ... return cleaned.startswith('0238')


    >>> recogniser = build_pteredactyl_recogniser(entity_type='SOUTHAMPTON_LANDLINE',
    ...                                           regex=r'(?:\\d[\\s-]?){11}',
    ...                                           check_function=check_soton_landline)
    """

    regex = re.compile(regex) if isinstance(regex, str) else regex
    return PteredactylRecogniser(
        entity_type=entity_type, regex=regex, check_function=check_function
    )


def fetch_pteredactyl_recogniser(entity_type: str) -> PteredactylRecogniser:
    if entity_type not in REGEX_ENTITIES.keys():
        raise MissingRegexRecogniserError(
            f"""Entity '{entity_type}' not found. Consider adding to the 'REGEX_ENTITIES' dictionary, or feeding a custom recogniser to regex_entities with:
 -> from pteredactyl.{Path(__file__).stem} import build_pteredactyl_recogniser
 -> recogniser = build_pteredactyl_recogniser(entity_type='{entity_type}', regex=r'some_regex', check_function=None) # or create a check_function with a single input argument, returning either True (correct match) or False"""
        )

    return build_pteredactyl_recogniser(
        entity_type=entity_type,
        regex=REGEX_ENTITIES[entity_type][0],
        check_function=REGEX_ENTITIES[entity_type][1],
    )


def build_regex_entity_recogniser_list(
    regex_entities: str | PteredactylRecogniser | Sequence[str | PteredactylRecogniser],
) -> list[PteredactylRecogniser]:
    """
    Build a list of custom regex PteredactylRecognisers.

    Args:
        regex_entities (list[str or PteredactylRecogniser]): A list of PteredactylRecogniser objects or strings referencing pre-built PteredactylRecognisers.

    Returns:
        list[PteredactylRecogniser]: A list of custom presidio EntityRecognizer objects.

    Example:
    >>> recognisers = build_regex_entity_recogniser_list(['NHS_NUMBER',
    ...                                                    'ENTITY_2',
    ...                                                    PteredactylRecogniser(entity_type='SOUTHAMPTON_LANDLINE',
    ...                                                                          regex=r'\\b((?:\\+44\\s?7\\d{3}|\\(?07\\d{3}\\)?)\\s?\\d{3}\\s?\\d{3}|\\(?01\\d{1,4}\\)?\\s?\\d{1,4}\\s?\\d{1,4})\\b',
    ...                                                                          check_function=check_so_landline
    ...                                                   ])
    """

    regex_entity_recognisers = []
    if type(regex_entities) in (str, PteredactylRecogniser):
        regex_entities = [regex_entities]

    for regex_entity in regex_entities:
        if isinstance(regex_entity, str):
            regex_entity_recognisers.append(
                fetch_pteredactyl_recogniser(entity_type=regex_entity)
            )
        else:
            regex_entity_recognisers.append(regex_entity)

    return regex_entity_recognisers


def rebuild_analyser_regex_recognisers(
    analyser: AnalyzerEngine, regex_entities: Sequence[str | PteredactylRecogniser]
) -> None:
    """
    Rebuilds the analyser's regex recognisers with the supplied list of regex entities.

    Args:
        analyser (AnalyzerEngine): The analyser to rebuild.
        regex_entities (list[str or PteredactylRecogniser]): The list of regex entities to use.

    Returns:
    """

    analyser.registry.remove_recognizer(PTEREDACTYL_RECOGNISER_NAME)
    pteredactyl_recognisers = build_regex_entity_recogniser_list(regex_entities)
    for recogniser in pteredactyl_recognisers:
        analyser.registry.add_recognizer(recogniser)



BASE_CONFIGURATION = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
        "AGE",
        "PHONE_NUMBER",
        "EMAIL_ADDRESS",
        "DATE_TIME",
        "DEVICE",
        "ZIP",
        "PROFESSION",
        "USERNAME",
        "ID",
    ],
    "LABELS_TO_IGNORE": ["O"],
    "SUB_WORD_AGGREGATION": "simple",
    "DATASET_TO_PRESIDIO_MAPPING": {
        "DATE": "DATE_TIME",
        "DOCTOR": "PERSON",
        "PATIENT": "PERSON",
        "HOSPITAL": "LOCATION",
        "MEDICALRECORD": "ID",
        "IDNUM": "ID",
        "ORGANIZATION": "ORGANIZATION",
        "ZIP": "ZIP",
        "PHONE": "PHONE_NUMBER",
        "USERNAME": "USERNAME",
        "STREET": "LOCATION",
        "PROFESSION": "PROFESSION",
        "COUNTRY": "LOCATION",
        "LOCATION-OTHER": "LOCATION",
        "FAX": "PHONE_NUMBER",
        "EMAIL": "EMAIL_ADDRESS",
        "STATE": "LOCATION",
        "DEVICE": "DEVICE",
        "ORG": "ORGANIZATION",
        "AGE": "AGE",
        "ID": "ID",
    },
    "MODEL_TO_PRESIDIO_MAPPING": {
        "PER": "PERSON",
        "PERSON": "PERSON",
        "LOC": "LOCATION",
        "ORG": "ORGANIZATION",
        "AGE": "AGE",
        "PATIENT": "PERSON",
        "HCW": "PERSON",
        "HOSPITAL": "LOCATION",
        "PATORG": "ORGANIZATION",
        "DATE": "DATE_TIME",
        "PHONE": "PHONE_NUMBER",
        "VENDOR": "ORGANIZATION",
        "ID": "ID",
        "FAC": "ZIP",
        "EMAIL": "EMAIL_ADDRESS",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}


def create_configuration(default_model_path, explanation):
    config = BASE_CONFIGURATION.copy()
    config["DEFAULT_MODEL_PATH"] = default_model_path
    config["DEFAULT_EXPLANATION"] = explanation
    return config


configuration = {
    "StanfordAIMI/stanford-deidentifier-base": create_configuration(
        "StanfordAIMI/stanford-deidentifier-base",
        "Identified as {} by the StanfordAIMI/stanford-deidentifier-base NER model",
    ),
    "StanfordAIMI/stanford-deidentifier-with-radiology-reports-and-i2b2": create_configuration(
        "StanfordAIMI/stanford-deidentifier-with-radiology-reports-and-i2b2",
        "Identified as {} by the StanfordAIMI/stanford-deidentifier-with-radiology-reports-and-i2b2 NER model",
    ),
    "lakshyakh93/deberta_finetuned_pii": create_configuration(
        "lakshyakh93/deberta_finetuned_pii",
        "Identified as {} by the lakshyakh93/deberta_finetuned_pii NER model",
    ),
    "urchade/gliner_multi_pii-v1": create_configuration(
        "urchade/gliner_multi_pii-v1",
        "Identified as {} by the urchade/gliner_multi_pii-v1 NER model",
    ),
    "beki/en_spacy_pii_distilbert": create_configuration(
        "beki/en_spacy_pii_distilbert",
        "Identified as {} by the beki/en_spacy_pii_distilbert NER model",
    ),
    "nikhilrk/de-identify": create_configuration(
        "nikhilrk/de-identify", "Identified as {} by the nikhilrk/de-identify NER model"
    ),
}



import re

import pytest

from pteredactyl.regex_check_functions import is_nhs_number
from pteredactyl.regex_entities import REGEX_ENTITIES

nhs_numbers = {
    "2345678909": True,
    "456-789-0124": True,
    "345 678 9017": True,
    "6789012346": True,
    "7890123450": True,
    "8901234564": True,
    "567 890 1230": True,
    9012345677: True,
    "0123456789": True,
    "1234567": False,
    "890 123 456": False,
    "1234567890": False,
    "567 890 1232": False,
    "9012345673": False,
    "0123456786": False,
}

uk_postcodes = {
    "B1A 1AA": True,
    "SW1A 1AA": True,
    "AAA 1AA": False,
    "E20 1EJ": True,
    "W1A 1AA": True,
    "ABCD 123": False,
    "EC1A 1BB": True,
    "W1": False,
    "B33 8TH": True,
    "N1A 1AA": True,
    "CR2 6XH": True,
    "DN55 1PT": True,
    "E10": False,
    "SW10 9RJ": True,
    "NW1E 2BU": True,
    "SE9 6RJ": True,
    "LS111AA": True,
}

email_addresses = {
    "valid.email@gmail.com": True,
    "user.name+tag+sorting@live.com": True,
    "user@sub.yahoo.com": True,
    "user.name@company.co.uk": True,
    "firstname.lastname@business.com": True,
    "user1234@gmail.com": True,
    "user@localhost": False,
    "user@website.web": True,
    "user@.com": False,
    "user@domain": False,
    "user@exam_ple.com": False,
}


def test_is_nhs_number():
    for nhs_number, expected_match in nhs_numbers.items():
        assert is_nhs_number(nhs_number) == expected_match


def test_is_uk_postcode():
    postcode_pattern = REGEX_ENTITIES["POSTCODE"][0]
    for postcode, expected_match in uk_postcodes.items():
        assert bool(re.search(postcode_pattern, postcode)) == expected_match


if __name__ == "__main__":
    pytest.main([__file__])



# read version from installed package
from importlib.metadata import version

__version__ = version("pteredactyl")

import warnings

import torch

from pteredactyl.defaults import (  # noqa: F401
    DEFAULT_ENTITIES,
    DEFAULT_NER_MODEL,
    DEFAULT_REGEX_ENTITIES,
    DEFAULT_SPACY_MODEL,
    show_defaults,
)
from pteredactyl.redactor import (  # noqa: F401
    analyse,
    anonymise,
    anonymise_df,
    create_analyser,
)
from pteredactyl.regex_entities import build_pteredactyl_recogniser  # noqa: F401

if not torch.cuda.is_available():
    warnings.warn(
        """
CUDA is not installed, so pteredactyl will use CPU rather than GPU.
    -> You can install CUDA 12.1 by running: pip3 install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu121 --upgrade
    -> Alternatively to select a compatible version visit: https://pytorch.org/get-started/locally/ and generate a pip3 installation command"""
    )



class MissingRegexRecogniserError(KeyError):
    """
    Exception raised when a regex recogniser is requested but not found in the supported regex_entities list.

    Attributes:
        message (str): The error message.
    """

    def __init__(
        self,
        message: str = "No regex settings could detected in pteredactyl.regex_entities",
    ):
        super().__init__(message)
        self.message = message



from typing import Any, NoReturn

from pteredactyl.mappings import configuration


def _get_config(model_path: str) -> dict[str, Any] | NoReturn:
    config = configuration.get(model_path)

    if config:
        return config

    else:
        available_list = [k for k in configuration.keys()]
        available = "\n -> " + (
            "\n -> ".join(available_list) if available_list else "None"
        )
        raise ValueError(
            f"Could not find config for '{model_path}' (please check pteredactyl.mappings). Available models: {available}"
        )
